import os
import sys

import dicom_parser

# cspell: ignore Wundefined afterloader angio autorefraction autosequence boli bscan cadsr ciexy ctdi ctdivol decf
# cspell: ignore deidentification dermoscopic diconde dicos dvhroi exif fiducials fluence fnumber frameof gpsdop hicpp
# cspell: ignore hpgl imager inms inua inus iods isop ivol ivus keratoconus keratometer keratometric kerma lensometry
# cspell: ignore lookat meterset metersets mrfov multiway nmpet nmtomo nolintbegin nolintend octb octz oecf olink
# cspell: ignore postamble powerline prcs presaturation radiofluoroscopic relaxivity reprojection rgblut rtdvh rtroi
# cspell: ignore sddn sdhn sdvn softcopy soundpath stereometric stowrs tlhc tomo tomosynthesis tomotherapeutic toric
# cspell: ignore tractography xaxrf

g_root_dir = os.path.dirname(__file__)
g_test_dir = os.path.join(os.path.dirname(g_root_dir), "test", "tu", "dicom")
g_header_extension = ".hpp"
g_source_extension = ".cpp"
g_test = "Test"

g_attribute_class_name = "Attribute"
g_sop_class_name = "Sop"


def generate_cspell_ignore(max_line_length: int = 120) -> str:
    cspell_words = {
        "bscan", "frameof", "isop", "nmpet", "nmtomo", "rtdvh", "afterloader", "angio", "autorefraction",
        "autosequence", "boli", "cadsr", "ciexy", "ctdi", "ctdivol", "decf", "deidentification", "dermoscopic",
        "diconde", "dicos", "dvhroi", "exif", "fiducials", "fluence", "fnumber", "gpsdop", "hicpp", "hpgl", "imager",
        "inms", "inua", "inus", "ivol", "ivus", "keratoconus", "keratometer", "keratometric", "kerma", "lensometry",
        "lookat", "meterset", "metersets", "mrfov", "multiway", "nolintbegin", "nolintend", "octb", "octz", "oecf",
        "postamble", "powerline", "prcs", "presaturation", "radiofluoroscopic", "relaxivity", "reprojection", "rgblut",
        "rtroi", "sddn", "sdhn", "sdvn", "softcopy", "soundpath", "stereometric", "stowrs", "tlhc", "tomo",
        "tomosynthesis", "tomotherapeutic", "toric", "tractography", "Wundefined", "xaxrf", "olink", "iods"
    }

    cspell_preamble = "// cspell: ignore"
    cspell_lines = cspell_preamble
    cspell_line_length = len(cspell_lines)

    for cspell_word in sorted(cspell_words):
        if cspell_line_length + len(cspell_word) >= max_line_length:
            new_line = cspell_preamble + " " + cspell_word
            cspell_line_length = len(new_line)
            cspell_lines += "\n" + new_line
        else:
            new_word = " " + cspell_word
            cspell_line_length += len(new_word)
            cspell_lines += new_word

    return cspell_lines


g_cspell_ignore = generate_cspell_ignore()


def find_sight_root_dir() -> str:
    """Find the SIGHT root directory"""
    sight_root_dir = g_root_dir
    while not os.path.exists(os.path.join(sight_root_dir, ".sight")):
        sight_root_dir = os.path.dirname(sight_root_dir)
    return sight_root_dir


g_script_path = __file__.replace(find_sight_root_dir(), "").replace("\\", "/")


def generate_preamble(includes: str, header: bool = True, namespace: str = None) -> str:
    """Generate the preamble for the generated files"""
    preamble = f"// This file was automatically generated by the '{g_script_path}' script. DO NOT EDIT !\n"
    preamble += f"// Used arguments: '{dicom_parser.g_arguments}'\n\n"

    if header:
        preamble += "#pragma once\n\n"

    if len(includes) > 0:
        preamble += f"{includes}\n\n"

    preamble += f"{g_cspell_ignore}\n\n"

    preamble += "// NOLINTBEGIN\n\n"

    if namespace is None:
        preamble += "namespace sight::data::dicom::ut\n{\n\n"
    else:
        preamble += f"namespace sight::data::dicom::{namespace}\n{{\n\n"

    return preamble


def generate_footer(namespace: str = None) -> str:
    """Generate the footer for the generated files"""
    if namespace is None:
        string = "} // namespace sight::data::dicom::ut\n\n"
    else:
        string = f"}} // namespace sight::data::dicom::{namespace}\n\n"

    string += "// NOLINTEND\n"

    return string


def tag_to_uint16(tag: dicom_parser.Tag) -> dicom_parser.Tag:
    """Convert a tag to a uint16_t hex string"""
    return dicom_parser.Tag(f"0x{tag.group.replace('x', 'F')}", f"0x{tag.element.replace('x', 'F')}")


def write_attribute_hpp() -> None:
    """Write the attribute.hpp file."""
    with open(os.path.join(g_root_dir, g_attribute_class_name + g_header_extension), mode='w',
              encoding='utf-8') as attribute_hpp:
        # Preamble
        attribute_hpp.write(
            generate_preamble(
                includes=(
                    "#include \"data/config.hpp\"\n\n"
                    "#include <cstdint>\n"
                    "#include <string_view>"
                ),
                header=True,
                namespace=g_attribute_class_name.lower()
            )
        )

        # enum Keyword
        attribute_hpp.write(
            f"enum class Keyword : std::uint16_t\n"
            "{\n"
            f"    {dicom_parser.g_invalid} = 0,\n"
        )

        for tag in sorted(dicom_parser.g_filtered_attributes):
            attribute = dicom_parser.g_filtered_attributes[tag]
            attribute_hpp.write(f"    {attribute.keyword},\n")

        attribute_hpp.write(
            "    _SIZE\n"
            "};\n\n"
        )

        # enum VR
        attribute_hpp.write(
            "enum class VR : std::uint8_t\n"
            "{\n"
            f"    {dicom_parser.g_invalid} = 0,\n"
        )

        for vr in sorted(dicom_parser.g_vrs):
            attribute_hpp.write(f"    {vr},\n")

        attribute_hpp.write(
            "    _SIZE\n"
            "};\n\n"
        )

        # enum VM
        attribute_hpp.write(
            "enum class VM : std::uint8_t\n"
            "{\n"
            f"    {dicom_parser.g_invalid} = 0,\n"
        )

        for vm in sorted(dicom_parser.g_vms):
            attribute_hpp.write(f"    {vm},\n")

        attribute_hpp.write(
            "    _SIZE\n"
            "};\n\n"
        )

        # IAttribute class
        attribute_hpp.write(
            f"struct DATA_CLASS_API I{g_attribute_class_name}\n"
            "{\n"
            f"    constexpr I{g_attribute_class_name}(\n"
            f"        const Keyword& keyword,\n"
            "        const std::uint16_t& group,\n"
            "        const std::uint16_t& element,\n"
            "        const std::string_view& name,\n"
            "        const VR& vr,\n"
            "        const VM& vm\n"
            "    ) noexcept :\n"
            "        m_keyword(keyword),\n"
            "        m_group(group),\n"
            "        m_element(element),\n"
            "        m_name(name),\n"
            "        m_vr(vr),\n"
            "        m_vm(vm)\n"
            "    {\n"
            "    }\n\n"
            f"    constexpr operator Keyword() const noexcept\n"
            "    {\n"
            "        return m_keyword;\n"
            "    }\n\n"
            f"    constexpr auto operator==(const I{g_attribute_class_name}& other) const noexcept\n"
            "    {\n"
            "       return m_keyword == other.m_keyword;\n"
            "    }\n\n"
            f"    const Keyword& m_keyword;\n"
            "    const std::uint16_t& m_group;\n"
            "    const std::uint16_t& m_element;\n"
            "    const std::string_view& m_name;\n"
            "    const VR& m_vr;\n"
            "    const VM& m_vm;\n"
            "};\n\n"
        )

        # Attribute class
        attribute_hpp.write(
            f"template<Keyword>\n"
            f"struct DATA_CLASS_API {g_attribute_class_name} final : public I{g_attribute_class_name}\n"
            "{\n"
            f"    constexpr {g_attribute_class_name}() noexcept :\n"
            f"        I{g_attribute_class_name}(s_keyword, s_group, s_element, s_name, s_vr, s_vm)\n"
            "    {\n"
            "    }\n\n"
            f"    SIGHT_DATA_API static const Keyword s_keyword;\n"
            "    SIGHT_DATA_API static const std::uint16_t s_group;\n"
            "    SIGHT_DATA_API static const std::uint16_t s_element;\n"
            "    SIGHT_DATA_API static const std::string_view s_name;\n"
            "    SIGHT_DATA_API static const VR s_vr;\n"
            "    SIGHT_DATA_API static const VM s_vm;\n"
            "};\n\n"
        )

        # attribute(attribute::Keyword) and attribute(group, element) methods
        attribute_hpp.write(
            f"/// @return a static const reference to an {g_attribute_class_name}\n"
            "/// @{\n"
            f"SIGHT_DATA_API const I{g_attribute_class_name}& get(Keyword keyword);\n"
            f"SIGHT_DATA_API const I{g_attribute_class_name}& get(std::uint16_t group, std::uint16_t element);\n"
            "/// @}\n\n"
        )

        attribute_hpp.write(generate_footer(namespace=g_attribute_class_name.lower()))


def write_attribute_cpp() -> None:
    """Write the Attribute.cpp file."""
    with open(os.path.join(g_root_dir, g_attribute_class_name + g_source_extension), mode='w',
              encoding='utf-8') as attribute_cpp:
        attribute_cpp.write(
            generate_preamble(
                includes=(
                    f"#include \"data/dicom/{g_attribute_class_name + g_header_extension}\"\n"
                    "#include <core/exceptionmacros.hpp>"
                ),
                header=False,
                namespace=g_attribute_class_name.lower()
            )
        )

        # get<Keyword::xxx>()
        attribute_cpp.write(
            f"template<Keyword keyword>\n"
            f"inline static const {g_attribute_class_name}<keyword>& get()\n"
            "{\n"
            f"    static constexpr {g_attribute_class_name}<keyword> instance;\n"
            "    return instance;\n"
            "}\n"
        )

        attribute_cpp.write(
            "\n//------------------------------------------------------------------------------\n\n"
        )

        # Attribute<Keyword::xxx>::s_yyy = zzz
        for tag in sorted(dicom_parser.g_filtered_attributes):
            attribute = dicom_parser.g_filtered_attributes[tag]
            uint16_tag = tag_to_uint16(tag)

            attribute_cpp.write(
                f"template<> SIGHT_DATA_API const Keyword {g_attribute_class_name}<Keyword::{attribute.keyword}>::s_keyword = Keyword::{attribute.keyword};\n"
                f"template<> SIGHT_DATA_API const std::uint16_t {g_attribute_class_name}<Keyword::{attribute.keyword}>::s_group = {uint16_tag.group};\n"
                f"template<> SIGHT_DATA_API const std::uint16_t {g_attribute_class_name}<Keyword::{attribute.keyword}>::s_element = {uint16_tag.element};\n"
                f"template<> SIGHT_DATA_API const std::string_view {g_attribute_class_name}<Keyword::{attribute.keyword}>::s_name = \"{attribute.name}\";\n"
                f"template<> SIGHT_DATA_API const VR {g_attribute_class_name}<Keyword::{attribute.keyword}>::s_vr = VR::{attribute.vr};\n"
                f"template<> SIGHT_DATA_API const VM {g_attribute_class_name}<Keyword::{attribute.keyword}>::s_vm = VM::{attribute.vm};\n"
            )

        attribute_cpp.write(
            "\n//------------------------------------------------------------------------------\n\n"
        )

        # get(Keyword)
        attribute_cpp.write(
            f"const I{g_attribute_class_name}& get(Keyword keyword)\n"
            "{\n"
            "    switch(keyword)\n"
            "    {\n"
        )

        for tag in sorted(dicom_parser.g_filtered_attributes):
            attribute = dicom_parser.g_filtered_attributes[tag]
            attribute_cpp.write(
                f"        case Keyword::{attribute.keyword}: return get<Keyword::{attribute.keyword}>();\n"
            )

        attribute_cpp.write(
            "        default:\n"
            f"            SIGHT_THROW(\"Unknown {g_attribute_class_name} Keyword: '\" << std::uint16_t(keyword) << \"'.\");\n"
            "    }\n"
        )

        attribute_cpp.write(
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        # get(group, element)
        attribute_cpp.write(
            f"const I{g_attribute_class_name}& get(std::uint16_t group, std::uint16_t element)\n"
            "{\n"
            "    // Special case for groups in the range 0x6000-0x60FF for attribute (60xx, xxxx)\n"
            "    if(group > 0x6000 && group <= 0x60FF)\n"
            "    {\n"
            "        group = 0x6000;\n"
            "    }\n\n"
            "    switch(group)\n"
            "    {\n"
        )

        # Order by group and then by element
        attribute_by_group_by_element = dict[str, dict[str, dicom_parser.Attribute]]()
        for tag in sorted(dicom_parser.g_filtered_attributes):
            attribute = dicom_parser.g_filtered_attributes[tag]
            uint16_tag = tag_to_uint16(tag)

            if uint16_tag.group not in attribute_by_group_by_element:
                attribute_by_group_by_element[uint16_tag.group] = dict[str, dicom_parser.Attribute]()

            attribute_by_group_by_element[uint16_tag.group][uint16_tag.element] = attribute

        # For each group
        for group in sorted(attribute_by_group_by_element):
            elements = attribute_by_group_by_element[group]

            attribute_cpp.write(
                f"        case {group}:\n"
                "        {\n"
                "            switch(element)\n"
                "            {\n"
            )

            for element in sorted(elements):
                attribute = elements[element]
                attribute_cpp.write(
                    f"            case {element}: return get<Keyword::{attribute.keyword}>();\n"
                )

            attribute_cpp.write(
                "            default:\n"
                "                SIGHT_THROW(\"Unknown element: '\" << element << \"'.\");\n"
                "            }\n"
                "        }\n\n"
            )

        attribute_cpp.write(
            "        default:\n"
            "            SIGHT_THROW(\"Unknown group: '\" << group << \"'.\");\n"
            "    }\n"
        )

        attribute_cpp.write(
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        attribute_cpp.write(generate_footer(namespace=g_attribute_class_name.lower()))


def write_sop_hpp() -> None:
    """Write the sop.hpp file."""
    with open(os.path.join(g_root_dir, g_sop_class_name + g_header_extension), mode='w', encoding='utf-8') as sop_hpp:
        # Preamble
        sop_hpp.write(
            generate_preamble(
                includes=(
                    "#include \"data/config.hpp\"\n"
                    "#include <cstdint>\n"
                    "#include <string_view>"
                ),
                header=True,
                namespace=g_sop_class_name.lower()
            )
        )

        # enum Keyword
        sop_hpp.write(
            f"enum class Keyword : std::uint8_t\n"
            "{\n"
            f"    {dicom_parser.g_invalid} = 0,\n"
        )

        for uid in sorted(dicom_parser.g_filtered_sops):
            sop_hpp.write(f"    {uid.keyword},\n")

        sop_hpp.write(
            "    _SIZE\n"
            "};\n\n"
        )

        # ISop
        sop_hpp.write(
            f"struct DATA_CLASS_API I{g_sop_class_name}\n"
            "{\n"
            f"    constexpr I{g_sop_class_name}(\n"
            f"        const Keyword& keyword,\n"
            "        const std::string_view& name,\n"
            "        const std::string_view& iud,\n"
            "        const std::string_view& iod\n"
            "    ) noexcept :\n"
            "        m_keyword(keyword),\n"
            "        m_name(name),\n"
            "        m_uid(iud),\n"
            "        m_iod(iod)\n"
            "    {\n"
            "    }\n\n"
            f"    constexpr operator Keyword() const noexcept\n"
            "    {\n"
            "        return m_keyword;\n"
            "    }\n\n"
            f"    constexpr auto operator==(const I{g_sop_class_name}& other) const noexcept\n"
            "    {\n"
            "       return m_keyword == other.m_keyword;\n"
            "    };\n\n"
            f"    const Keyword& m_keyword;\n"
            "    const std::string_view& m_name;\n"
            "    const std::string_view& m_uid;\n"
            "    const std::string_view& m_iod;\n"
            "};\n\n"
        )

        # Sop
        sop_hpp.write(
            f"template<Keyword>\n"
            f"struct DATA_CLASS_API {g_sop_class_name} final : public I{g_sop_class_name}\n"
            "{\n"
            f"    constexpr {g_sop_class_name}() noexcept :\n"
            f"        I{g_sop_class_name}(s_keyword, s_name, s_uid, s_iod)\n"
            "    {\n"
            "    }\n\n"
            f"    SIGHT_DATA_API static const Keyword s_keyword;\n"
            "    SIGHT_DATA_API static const std::string_view s_name;\n"
            "    SIGHT_DATA_API static const std::string_view s_uid;\n"
            "    SIGHT_DATA_API static const std::string_view s_iod;\n"
            "};\n\n"
        )

        # get(Keyword) and keyword(uid)
        sop_hpp.write(
            f"/// @return a static const reference to an {g_sop_class_name}\n"
            "/// @{\n"
            f"SIGHT_DATA_API const I{g_sop_class_name}& get(Keyword keyword);\n"
            f"SIGHT_DATA_API const I{g_sop_class_name}& get(const std::string_view& uid);\n"
            "/// @}\n\n"
            "/// convert a uid directly to a keyword\n"
            "SIGHT_DATA_API const Keyword& keyword(const std::string_view& uid) noexcept;\n\n"
        )

        sop_hpp.write(generate_footer(namespace=g_sop_class_name.lower()))


def write_sop_cpp() -> None:
    """Write the Sop.cpp file."""
    with open(os.path.join(g_root_dir, g_sop_class_name + g_source_extension), mode='w', encoding='utf-8') as sop_cpp:
        sop_cpp.write(
            generate_preamble(
                includes=(
                    f"#include \"data/dicom/{g_sop_class_name + g_header_extension}\"\n"
                    "#include <core/exceptionmacros.hpp>\n"
                    "#include <unordered_map>"
                ),
                header=False,
                namespace=g_sop_class_name.lower()
            )
        )

        # get<Keyword::xxx>()
        sop_cpp.write(
            f"template<Keyword keyword>\n"
            f"inline static const {g_sop_class_name}<keyword>& get()\n"
            "{\n"
            f"    static constexpr {g_sop_class_name}<keyword> instance;\n"
            "    return instance;\n"
            "}\n"
        )

        sop_cpp.write(
            "\n//------------------------------------------------------------------------------\n\n"
        )

        # Sop<Keyword::xxx>::s_yyy = zzz
        for uid in sorted(dicom_parser.g_filtered_sops):
            sop = dicom_parser.g_filtered_sops[uid]
            iod = dicom_parser.g_filtered_iods[sop.iod.keyword]
            sop_cpp.write(
                f"template<> SIGHT_DATA_API const Keyword {g_sop_class_name}<Keyword::{uid.keyword}>::s_keyword = Keyword::{uid.keyword};\n"
                f"template<> SIGHT_DATA_API const std::string_view {g_sop_class_name}<Keyword::{uid.keyword}>::s_name = \"{uid.name}\";\n"
                f"template<> SIGHT_DATA_API const std::string_view {g_sop_class_name}<Keyword::{uid.keyword}>::s_uid = \"{uid.value}\";\n"
                f"template<> SIGHT_DATA_API const std::string_view {g_sop_class_name}<Keyword::{uid.keyword}>::s_iod = \"{iod.name}\";\n"
            )

        sop_cpp.write(
            "\n//------------------------------------------------------------------------------\n\n"
        )

        # get(Keyword)
        sop_cpp.write(
            f"const I{g_sop_class_name}& get(Keyword keyword)\n"
            "{\n"
            "    switch(keyword)\n"
            "    {\n"
        )

        for uid in sorted(dicom_parser.g_filtered_sops):
            sop_cpp.write(
                f"        case Keyword::{uid.keyword}: return get<Keyword::{uid.keyword}>();\n"
            )

        sop_cpp.write(
            "        default:\n"
            f"            SIGHT_THROW(\"Unknown {g_sop_class_name} Keyword: '\" << std::uint8_t(keyword) << \"'.\");\n"
            "    }\n"
        )

        sop_cpp.write(
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        # get(uid)
        sop_cpp.write(
            f"const I{g_sop_class_name}& get(const std::string_view& uid)\n"
            "{\n"
            f"    static const std::unordered_map<std::string_view, const I{g_sop_class_name}&> {g_sop_class_name.upper()}_MAP {{\n"
        )

        for uid in sorted(dicom_parser.g_filtered_sops):
            sop_cpp.write(
                f"        {{{g_sop_class_name}<Keyword::{uid.keyword}>::s_uid, get<Keyword::{uid.keyword}>()}},\n"
            )

        sop_cpp.write(
            "    };\n\n"
            f"    return {g_sop_class_name.upper()}_MAP.at(uid);\n"
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        sop_cpp.write(
            f"const Keyword& keyword(const std::string_view& uid) noexcept\n"
            "{\n"
            "    try\n"
            "    {\n"
            f"        return get(uid).m_keyword;\n"
            "    }\n"
            "    catch(...)\n"
            "    {\n"
            "        static const auto invalid = Keyword::INVALID;\n"
            "        return invalid;\n"
            "    }\n"
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        sop_cpp.write(generate_footer(namespace=g_sop_class_name.lower()))


def write_attribute_test_hpp() -> None:
    """Write the AttributeTest.hpp file."""
    with open(
            os.path.join(g_test_dir, g_attribute_class_name + g_test + g_header_extension),
            mode='w',
            encoding='utf-8'
    ) as attribute_test_hpp:
        attribute_test_hpp.write(
            generate_preamble(
                includes=(
                    "#include <cppunit/extensions/HelperMacros.h>"
                ),
                header=True
            )
        )

        # Test suite / methods
        attribute_test_hpp.write(
            f"class {g_attribute_class_name}{g_test} : public CPPUNIT_NS::TestFixture\n"
            "{\n"
            f"    CPPUNIT_TEST_SUITE({g_attribute_class_name}{g_test});\n"
            f"    CPPUNIT_TEST({g_attribute_class_name.lower()}{g_test});\n"
            "    CPPUNIT_TEST_SUITE_END();\n\n"
            "public:\n"
            "    void setUp() override;\n"
            "    void tearDown() override;\n\n"
            f"    static void {g_attribute_class_name.lower()}{g_test}();\n"
            "};\n\n"
        )

        attribute_test_hpp.write(generate_footer())


def write_attribute_test_cpp() -> None:
    """Write the AttributeTest.cpp file."""
    with open(
            os.path.join(g_test_dir, g_attribute_class_name + g_test + g_source_extension),
            mode='w',
            encoding='utf-8'
    ) as attribute_test_cpp:
        attribute_test_cpp.write(
            generate_preamble(
                includes=(
                    f"#include \"{g_attribute_class_name}{g_test}{g_header_extension}\"\n\n"
                    f"#include <data/dicom/{g_attribute_class_name}{g_header_extension}>\n\n"
                    f"CPPUNIT_TEST_SUITE_REGISTRATION(sight::data::dicom::ut::{g_attribute_class_name}{g_test});"
                ),
                header=False
            )
        )

        attribute_test_cpp.write(
            f"void {g_attribute_class_name}{g_test}::setUp()\n"
            "{\n"
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
            f"void {g_attribute_class_name}{g_test}::tearDown()\n"
            "{\n"
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        attribute_test_cpp.write(
            f"void {g_attribute_class_name}{g_test}::{g_attribute_class_name.lower()}{g_test}()\n"
            "{\n"
        )

        for tag in sorted(dicom_parser.g_filtered_attributes):
            attribute = dicom_parser.g_filtered_attributes[tag]
            uint16_tag = tag_to_uint16(tag)

            attribute_test_cpp.write(
                f"    CPPUNIT_ASSERT_EQUAL({g_attribute_class_name.lower()}::{g_attribute_class_name}<{g_attribute_class_name.lower()}::Keyword::{attribute.keyword}>::s_keyword, {g_attribute_class_name.lower()}::get({uint16_tag.group}, {uint16_tag.element}).m_keyword);\n"
                f"    CPPUNIT_ASSERT_EQUAL({g_attribute_class_name.lower()}::{g_attribute_class_name}<{g_attribute_class_name.lower()}::Keyword::{attribute.keyword}>::s_keyword, {g_attribute_class_name.lower()}::get({g_attribute_class_name.lower()}::Keyword::{attribute.keyword}).m_keyword);\n"
            )

        attribute_test_cpp.write(
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        attribute_test_cpp.write(generate_footer())


def write_sop_test_hpp() -> None:
    """Write the SopTest.hpp file."""
    with open(
            os.path.join(g_test_dir, g_sop_class_name + g_test + g_header_extension),
            mode='w',
            encoding='utf-8'
    ) as sop_test_hpp:
        sop_test_hpp.write(
            generate_preamble(
                includes=(
                    "#include <cppunit/extensions/HelperMacros.h>"
                ),
                header=True
            )
        )

        # Test suite / methods
        sop_test_hpp.write(
            f"class {g_sop_class_name}{g_test} : public CPPUNIT_NS::TestFixture\n"
            "{\n"
            f"    CPPUNIT_TEST_SUITE({g_sop_class_name}{g_test});\n"
            f"    CPPUNIT_TEST({g_sop_class_name.lower()}{g_test});\n"
            "    CPPUNIT_TEST_SUITE_END();\n\n"
            "public:\n"
            "    void setUp() override;\n"
            "    void tearDown() override;\n\n"
            f"    static void {g_sop_class_name.lower()}{g_test}();\n"
            "};\n\n"
        )

        sop_test_hpp.write(generate_footer())


def write_sop_test_cpp() -> None:
    """Write the AttributeTest.cpp file."""
    with open(
            os.path.join(g_test_dir, g_sop_class_name + g_test + g_source_extension),
            mode='w',
            encoding='utf-8'
    ) as sop_test_cpp:
        sop_test_cpp.write(
            generate_preamble(
                includes=(
                    f"#include \"{g_sop_class_name}{g_test}{g_header_extension}\"\n\n"
                    f"#include <data/dicom/{g_sop_class_name}{g_header_extension}>\n\n"
                    f"CPPUNIT_TEST_SUITE_REGISTRATION(sight::data::dicom::ut::{g_sop_class_name}{g_test});"
                ),
                header=False
            )
        )

        sop_test_cpp.write(
            f"void {g_sop_class_name}{g_test}::setUp()\n"
            "{\n"
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
            f"void {g_sop_class_name}{g_test}::tearDown()\n"
            "{\n"
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        sop_test_cpp.write(
            f"void {g_sop_class_name}{g_test}::{g_sop_class_name.lower()}{g_test}()\n"
            "{\n"
        )

        for uid in sorted(dicom_parser.g_filtered_sops):
            sop_test_cpp.write(
                f"    CPPUNIT_ASSERT_EQUAL({g_sop_class_name.lower()}::{g_sop_class_name}<{g_sop_class_name.lower()}::Keyword::{uid.keyword}>::s_keyword, {g_sop_class_name.lower()}::get({g_sop_class_name.lower()}::Keyword::{uid.keyword}).m_keyword);\n"
                f"    CPPUNIT_ASSERT_EQUAL({g_sop_class_name.lower()}::{g_sop_class_name}<{g_sop_class_name.lower()}::Keyword::{uid.keyword}>::s_keyword, {g_sop_class_name.lower()}::get(\"{uid.value}\").m_keyword);\n"
            )

        sop_test_cpp.write(
            "}\n\n"
            "//------------------------------------------------------------------------------\n\n"
        )

        sop_test_cpp.write(generate_footer())


def main():
    """Main function"""

    result = dicom_parser.main()

    # Attribute
    write_attribute_hpp()
    write_attribute_cpp()

    # Attribute tests
    write_attribute_test_hpp()
    write_attribute_test_cpp()

    # Sop
    write_sop_hpp()
    write_sop_cpp()

    # Sop tests
    write_sop_test_hpp()
    write_sop_test_cpp()

    return result


if __name__ == "__main__":
    sys.exit(main())



#!/bin/python3
"""
Generate a dash docset from a doxygen documentation.
"""

# cspell: disable
import json
import os
from pathlib import Path
import re
import shutil
import sqlite3
import sys

import bs4

CFG = dict()
REPO_NAMES = dict()

# Global regexes we don't want to recompile every single time we parse a file
CLASS_FILE_RE = re.compile(r'class([a-zA-Z_][a-zA-Z0-9_]*)_1_1([a-zA-Z_][a-zA-Z0-9_]*)\.html')
CLASS_RE = re.compile('sight: (.+) Class Reference')
STRUCT_FILE_RE = re.compile(r'struct([a-zA-Z_][a-zA-Z0-9_]*)_1_1([a-zA-Z_][a-zA-Z0-9_]*)\.html')
STRUCT_RE = re.compile('sight: (.+) Struct Reference')
NAMESPACE_FILE_RE = re.compile(r'namespace.+\.html')
NAMESPACE_RE = re.compile('sight: ([a-zA-Z_][a-zA-Z0-9_:]*) Namespace Reference')
SRV_RE = re.compile('sight: ([a-zA-Z_][a-zA-Z0-9_]*::(?:[a-zA-Z_][a-zA-Z0-9_]*::)*(S[A-Z0-9][a-zA-Z0-9_]*)) Class Reference')
BAD__SRV_RE = re.compile('sight: ([a-zA-Z_][a-zA-Z0-9_]*::(?:[a-zA-Z_][a-zA-Z0-9_]*::)*([A-Z0-9][a-zA-Z0-9_]*)) Class Reference')
OBJ_RE = re.compile('sight: ([a-zA-Z_][a-zA-Z0-9_]*::(?:[a-zA-Z_][a-zA-Z0-9_]*::)*([A-Z0-9][a-zA-Z0-9_]*)) Class Reference')
IFACE_RE = re.compile('sight: ([a-zA-Z_][a-zA-Z0-9_]*::(?:[a-zA-Z_][a-zA-Z0-9_]*::)*(I[A-Z0-9][a-zA-Z0-9_]*|base)) Class Reference')
EXCEPT_RE = re.compile('sight: ([a-zA-Z_][a-zA-Z0-9_]*::(?:[a-zA-Z_][a-zA-Z0-9_]*::)*([A-Z0-9][a-zA-Z0-9_]*)) Struct Reference')

# Regexes of the files to skip
FILE_SKIP_RE = [
    re.compile('pages.html'),
    re.compile(r'dir_.+\.html'),
    re.compile('.+_source.html')
]

def bootstrap_docset():
    """
    Create the skeleton for the docset, i.e. the directory structure along with the SQLite database. Return the SQLite
    database connection.
    """
    # Create the directory structure.
    Path('./sight.docset/Contents/Resources').mkdir(parents=True, exist_ok=True)

    # Then, create the SQLite database
    db = Path('./sight.docset/Contents/Resources/docSet.dsidx')
    if db.exists():
        os.remove(str(db))
    conn_ = sqlite3.connect(str(db))
    cur = conn_.cursor()
    cur.execute('CREATE TABLE searchIndex(id INTEGER PRIMARY KEY, name TEXT, type TEXT, path TEXT);')
    cur.execute('CREATE UNIQUE INDEX anchor ON searchIndex (name, type, path);')
    conn_.commit()
    return conn_

def gather_sources():
    """
    Return a list containing the paths to all interesting HTML files contained at the root of the Doxygen html
    directory. We're not interested in what's in the subdirectories.
    """
    files = list()
    for _, _, dir_files in os.walk('./html/'):
        files += [f for f in dir_files if f.endswith('.html')]
    return files

def parse_related_pages():
    """
    Parse the 'pages.html' doxygen file and generate the list of related pages.
    """
    pages = list()
    html = open(os.path.join('./html', 'pages.html'), encoding="utf8").read()
    soup = bs4.BeautifulSoup(html, "html.parser")
    table = soup.find("table", class_="directory")
    for cell in table.find_all("tr"):
        page_name = cell.td.a.string
        page_link = cell.td.a.get('href')
        pages.append((page_name, "Guide", page_link))
    return pages

# TODO: strip the leading repository path from the HTML file to get rid of the ugly full paths ?
def file_repo(f_soup):
    """
    Return the name of the repository that a particular documentation file was generated from, or None if not possible.
    """
    lists = f_soup.find_all('ul')
    if lists:
        file_path = lists[-1].li.get_text()
        for repo in CFG['repositories']:
            candidates = [repo for repo in CFG['repositories'] if file_path.startswith(repo)]
            if candidates:
                res = max(candidates, key=len)
                return REPO_NAMES[res]
    return None

def parse_file(f_):
    """
    Parse a HTML file and return a (potentially empty) list of 3-tuples to add to the SQLite database.
    """
    # Doxygen names the documentation files in a friendly manner, which means we can guess what is inside from the file
    # name, and ignore files that we know we don't care about. This script currently looks for files containing classes
    # or structs.
    new_entries = list()
    # Some files are of no interest to us and can be skipped
    if any(map(lambda regexp: regexp.match(f_), FILE_SKIP_RE)):
        return new_entries
    try:
        html = open(os.path.join('./html', f_), encoding="utf8").read()
        soup = bs4.BeautifulSoup(html, "html.parser")
        inherits_base = soup.find(class_='inherit_header pub_methods_classfwServices_1_1base')
        inherits_object = soup.find(class_='inherit_header pub_methods_classfwData_1_1Object')
        inherits_exception = soup.find(class_='inherit_header pub_methods_classfwCore_1_1Exception')

        item_type_re = {
            "Service": SRV_RE,
            "Object": OBJ_RE,
            "Interface": IFACE_RE,
            "Exception": EXCEPT_RE
        }
        def is_item_type(soup, ty_str):
            """
            Test if the HTML contained in the supplied soup describes and element of the specified type based on the
            doxygen page title. Accepted types are 'Service', 'Object', 'Interface' and 'Exception'. If true, return an
            entry to add to the sqlite DB, else return None.
            """
            title = soup.title.get_text()
            match = item_type_re[ty_str].search(title)
            if match:
                path = match.group(1)
                repo = file_repo(soup)
                if repo is not None:
                    path = path + " ({})".format(repo)
                return (path, ty_str, f_)
            return None

        def is_bad_service(soup):
            """
            Test if the HTML contained in the supplied soup describes a service, with more lenient rules regarding
            the name of the service. If true, print a warning regarding the service name and return an entry to add to
            the sqlite DB, otherwise return None.
            """
            title = soup.title.get_text()
            match = BAD__SRV_RE.search(title)
            if match:
                path = match.group(1)
                srv = match.group(2)
                repo = file_repo(soup)
                if repo is not None:
                    path = path + " ({})".format(repo)
                print("Warning: service {} has non compliant name (no S prefix)".format(srv))
                return (path, "Service", f_)
            return None

        file_type_re = {
            "Class": CLASS_RE,
            "Namespace": NAMESPACE_RE,
            "Struct": STRUCT_RE,
        }
        def is_file_type(soup, ty_str):
            """
            Test if the HTML contained in the supplied soup describes and element of the specified type based on the
            doxygen page title. Accepted types are 'Class', 'Namespace', and 'Struct'. If true, return an
            entry to add to the sqlite DB, else return None.
            """
            title = soup.title.get_text()
            match = file_type_re[ty_str].search(title)
            if match:
                struct_ = match.group(1)
                return (struct_, ty_str, f_)
            return None

        if CLASS_FILE_RE.match(f_):
            # We know the file contains a class, find what kind of class
            class_triple = is_file_type(soup, 'Class')
            if class_triple is None:
                return new_entries
            class_name = class_triple[0]
            if inherits_base:
                # The class inherits base, it can be a service or an interface
                triple = is_item_type(soup, 'Interface')
                if triple is not None:
                    new_entries.append(triple)
                else:
                    # Not an interface, probably a service
                    triple = is_item_type(soup, 'Service')
                    if triple is not None:
                        new_entries.append(triple)
                    else:
                        triple = is_bad_service(soup)
                        if triple is not None:
                            new_entries.append(triple)
                        else:
                            print("Warning: unexepected behaviour for class {} while parsing file {}".format(class_name, f_))
            elif class_name == "fwData::Object":
                # Special case, Object is not an actual data.
                new_entries.append((class_name, "Class", f_))
            elif inherits_object:
                # Not a service and inherits fwData::Object, this class is probably a data.
                triple = is_item_type(soup, 'Object')
                if triple is not None:
                    new_entries.append(triple)
            elif class_name == "fwCore::Exception":
                # Special case for fwCore::Exception
                new_entries.append((class_name, "Exception", f_))
            elif inherits_exception:
                # Inherits an exception type, this is probably an exception
                # TODO: I'm pretty sure this won't catch all exceptions in the codebase
                triple = is_item_type(soup, 'Exception')
                if triple is not None:
                    new_entries.append(triple)
            else:
                # Plain old class
                new_entries.append(class_triple)
        elif STRUCT_FILE_RE.match(f_):
            # We know the file contains a struct, find what kind of struct
            struct_triple = is_file_type(soup, 'Struct')
            if struct_triple is None:
                return new_entries
            new_entries.append(struct_triple)
            if inherits_exception:
                # Inherits an exception type, this is probably an exception
                # TODO: I'm pretty sure this won't catch all exceptions in the codebase
                triple = is_item_type(soup, 'Exception')
                if triple is not None:
                    new_entries.append(triple)
        elif NAMESPACE_FILE_RE.match(f_):
            # We know the file contains a namespace, find what kind of namespace (i.e. Bundle, Library, regular
            # namespace...)
            namespace_triple = is_file_type(soup, 'Namespace')
            if namespace_triple is None:
                return new_entries
            namespace_name = namespace_triple[0]
            if namespace_name in CFG['srclibs']:
                new_entries.append((namespace_name, "Library", f_))
            elif namespace_name in CFG['bundles']:
                # There is no 'Bundle' entry type, unfortunately. Component, Package or Module would be suitable
                # replacements. I chose Package.
                new_entries.append((namespace_name, "Package", f_))
            else:
                new_entries.append(namespace_triple)
    except UnicodeDecodeError:
        print('The file ' + f_ + ' is not valid UTF-8')
    except FileNotFoundError:
        # Occurs for files in the search subdirectory, it's OK, we don't care about those
        pass
    return new_entries

def populate_db(conn_, services):
    """
    Fill the sqlite database with the supplied list of (name, entry_type, file_path) triples.
    """
    cur = conn_.cursor()
    for triple in services:
        try:
            cur.execute("INSERT OR IGNORE INTO searchIndex(name, type, path) VALUES (?, ?, ?);", triple)
        except Exception:
            print("Error inserting " + str(triple))
    conn_.commit()

def copy_files():
    """
    Copy the doxygen HTML files into the docset destination.
    """
    try:
        shutil.copytree('./html', './sight.docset/Contents/Resources/Documents')
    except shutil.Error as err:
        errors = err.args[0]
        print("Warning: some files were not copied correctly. The generated docset might be incomplete.")
        for src, _, why in errors:
            print("File '" + src + "' was not copied correctly. Reason: " + why)

def main():
    """
    Builds the dash docset.
    """
    global CFG
    global REPO_NAMES
    try:
        CFG = json.loads(open('./projects.json', encoding="utf8").read())
    except (OSError, json.JSONDecodeError) as err:
        print("Error loading configuration file: " + str(err))
        return
    REPO_NAMES = {repo: Path(repo).parent.name if Path(repo).name == "src" else Path(repo).name for repo in CFG['repositories']}
    if CFG is None:
        sys.exit(1)
    conn = bootstrap_docset()
    html_files = gather_sources()
    entries = list()
    for f in html_files:
        f_entries = parse_file(f)
        if f_entries:
            entries += f_entries
    entries += parse_related_pages()
    populate_db(conn, entries)
    copy_files()
    conn.close()

if __name__ == '__main__':
    main()



import argparse
import fnmatch
import os
import re
import sys
import xml.etree.ElementTree
from dataclasses import dataclass
from urllib.request import urlretrieve


# cspell: ignore Wundefined afterloader angio autorefraction autosequence boli bscan cadsr ciexy ctdi ctdivol decf
# cspell: ignore deidentification dermoscopic diconde dicos dvhroi exif fiducials fluence fnumber frameof gpsdop hicpp
# cspell: ignore hpgl imager inms inua inus iods isop ivol ivus keratoconus keratometer keratometric kerma lensometry
# cspell: ignore lookat meterset metersets mrfov multiway nmpet nmtomo nolintbegin nolintend octb octz oecf olink
# cspell: ignore postamble powerline prcs presaturation radiofluoroscopic relaxivity reprojection rgblut rtdvh rtroi
# cspell: ignore sddn sdhn sdvn softcopy soundpath stereometric stowrs tlhc tomo tomosynthesis tomotherapeutic toric
# cspell: ignore tractography xaxrf

@dataclass(eq=True, order=True, frozen=True)
class Tag:
    group: str
    element: str


@dataclass(eq=True, order=True, frozen=True)
class Attribute:
    tag: Tag
    name: str
    keyword: str
    vr: str
    vm: str
    description: str


@dataclass(eq=True, order=True, frozen=True)
class AttributeElement:
    attribute: Attribute
    type: str
    description: str
    attribute_elements: list
    origin: str


@dataclass(eq=True, order=True, frozen=True)
class Module:
    name: str
    keyword: str
    attribute_elements: list[AttributeElement]


@dataclass(eq=True, order=True, frozen=True)
class ModuleElement:
    module: Module
    usage: str
    condition: str


@dataclass(eq=True, order=True, frozen=True)
class Iod:
    name: str
    keyword: str
    module_elements: list[ModuleElement]
    functional_group_elements: list[ModuleElement]


@dataclass(eq=True, order=True, frozen=True)
class Uid:
    value: str
    name: str
    keyword: str
    type: str


@dataclass(eq=True, order=True, frozen=True)
class Sop:
    uid: Uid
    iod: Iod


# Program arguments
g_parser = None
g_arguments = None
g_source_url = None
g_mandatory_tags = list[Tag]()
g_sop_list = list[str]()

# Global XML namespaces
g_namespaces = {
    "xml": "http://www.w3.org/XML/1998/namespace",
    "docbook": "http://docbook.org/ns/docbook"
}

g_invalid = "INVALID"

# To avoid recompiling regexes
g_snake_regex1 = re.compile(r"((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")
g_keyword_regex = re.compile(r"[\W_]+")
g_tag_regex = re.compile(r"\(\s*([0-9A-Fx]{4})\s*,\s*([0-9A-Fx]{4})\s*\)")
g_uid_regex = re.compile(r"(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*")
g_uid_wildcard_regex = re.compile(r"[0-9.*?]+")

# Various cache to speedup parsing
g_documents_cache = dict[str, xml.etree.ElementTree.ElementTree]()
g_iods_cache = dict[str, Iod]()
g_modules_cache = dict[str, Module]()
g_attribute_elements_cache = dict[str, list[AttributeElement]]()

# Parsing results
g_attributes = dict[Tag, Attribute]()
g_vrs = set[str]()
g_vms = set[str]()
g_uids = dict[str, Uid]()
g_uid_types = set[str]()
g_filtered_sops = dict[Uid, Sop]()
g_filtered_iods = dict[str, Iod]()
g_filtered_modules = dict[str, Module]()
g_filtered_attributes = dict[Tag, Attribute]()
g_filtered_attribute_types = set[str]()


def check_mandatory_tags(sop: Sop) -> bool:
    """Check if a SOP is matching the tag filters"""

    # Check if we have filters
    if g_mandatory_tags is None or len(g_mandatory_tags) == 0:
        return True

    # Check if the SOP has a matching tag
    iod = sop.iod

    def check_module_element(a_module_element: ModuleElement) -> bool:
        module = a_module_element.module
        if module is not None and module.attribute_elements is not None:

            def check_attribute_elements(attribute_elements: list[AttributeElement], path: list[str]) -> bool:
                for attribute_element in attribute_elements:
                    if attribute_element.attribute.tag in g_mandatory_tags:
                        return True

                    # Check sub-attribute elements
                    sub_attribute_elements = attribute_element.attribute_elements
                    if (
                            sub_attribute_elements is not None
                            and len(sub_attribute_elements) > 0
                            and attribute_element.origin not in path
                    ):
                        # Avoid infinite recursion
                        new_path = path + [attribute_element.origin]

                        # Recurse
                        if check_attribute_elements(sub_attribute_elements, new_path):
                            return True
                return False

            if check_attribute_elements(module.attribute_elements, []):
                return True

        return False

    for module_element in iod.module_elements:
        if check_module_element(module_element):
            return True

    for module_element in iod.functional_group_elements:
        if check_module_element(module_element):
            return True

    return False


def open_document(part_name: str) -> xml.etree.ElementTree:
    """Open a document and cache it"""

    # Check the cache
    if part_name in g_documents_cache:
        return g_documents_cache[part_name]

    # Download the file if needed
    filename = f"{part_name}.xml"
    if not os.path.isfile(filename):
        url = f"{g_source_url}/{part_name}/{filename}"
        print(f"Downloading {url}...")
        urlretrieve(url, filename)

    print(f"Opening {filename}...")
    with open(file=filename, mode="r", encoding="utf-8") as document_file:
        document = xml.etree.ElementTree.parse(document_file)
        g_documents_cache[part_name] = document
        return document


def to_keyword(name: str) -> str:
    """Convert a name to a PascalCase"""
    # Remove any non-alphanumeric character
    result = g_keyword_regex.sub('_', name)

    # Capitalise each word (But keep already uppercase characters)
    return "".join(word[0].upper() + word[1:] for word in result.split('_'))


def xml_element_to_string(element: xml.etree.ElementTree.Element) -> str:
    """Convert an element to a string"""
    emphasis = element.find("./docbook:emphasis", g_namespaces)

    if emphasis is not None:
        element = emphasis

    string = element.text or ""
    string = string.replace(u"\u200b", "").replace(u"\u00b5", "u")
    return string.strip()


def xml_element_to_tag(element: xml.etree.ElementTree.Element) -> Tag:
    """Convert a tag element to a group and element string"""

    # Get the tag string
    tag_string = xml_element_to_string(element)

    # Clean the tag
    tag_string = tag_string.replace("(", "").replace(")", "").replace("x", "0")

    # Split the group and the element
    split_tag = tag_string.split(",")

    return Tag(
        group=split_tag[0].strip(),
        element=split_tag[1].strip()
    )


def xml_element_to_vr(element: xml.etree.ElementTree.Element) -> str:
    """Convert a vr element to a vr string"""

    # Get the vr string
    vr_string = xml_element_to_string(element)

    # If there is no vr, return "INVALID"
    if len(vr_string) == 0 or vr_string.find("See") != -1:
        return g_invalid

    # If the vr is a list, split and reconstruct it
    split_vr = [v.strip() for v in vr_string.split(" or ")]
    joined_vr = "_".join(sorted(split_vr))

    return joined_vr


def xml_element_to_vm(element: xml.etree.ElementTree.Element) -> str:
    """Convert a vm element to a vm string"""

    # Get the vm string
    vm_string = xml_element_to_string(element).lower()

    # If there is no vm, return "INVALID"
    if len(vm_string) == 0:
        return g_invalid

    # If the vm can be a range or 1, take the range
    if vm_string.find("1-n or 1") != -1:
        vm_string = "1-n"

    # Normalize all vm as ranges
    split_vm = vm_string.split("-")
    vm = "MIN_" + split_vm[0] + "_MAX_"

    if len(split_vm) >= 2:
        vm += split_vm[1]
    else:
        vm += split_vm[0]

    return vm.upper()


def xml_element_to_usage(element: xml.etree.ElementTree.Element) -> tuple[str, str]:
    """Convert a usage element to a "usage" / "condition" string pair"""
    pair = xml_element_to_string(element).split("-")
    if len(pair) == 1:
        return pair[0].strip(), ""
    elif len(pair) > 2:
        return pair[0].strip(), "-".join(pair[1:]).strip()
    else:
        return pair[0].strip(), pair[1].strip()


def xml_element_to_uid(element: xml.etree.ElementTree.Element) -> str:
    """Convert an element to a uid"""
    uid = xml_element_to_string(element)

    if g_uid_regex.match(uid) is None:
        raise ValueError(f"Invalid UID: {uid}")

    return uid


def parse_attributes() -> tuple[dict[Tag, Attribute], set[str], set[str]]:
    """Parse attributes from DICOM docbook files"""

    # List all source files and their tables
    sources = [
        ("part07", {"table_E.1-1"}),
        ("part06", {"table_7-1", "table_8-1", "table_6-1"})
    ]

    attributes = dict[Tag, Attribute]()
    vrs = set[str]()
    vms = set[str]()

    # For each source files
    for source in sources:
        document = open_document(source[0])

        # For each table
        for table_id in source[1]:

            # Find the table
            table = document.find(f".//*[@xml:id=\"{table_id}\"]", g_namespaces)

            for row in table.iterfind("./docbook:tbody/docbook:tr", g_namespaces):
                columns = row.findall("./docbook:td/docbook:para", g_namespaces)

                if len(columns) > 5:
                    description = xml_element_to_string(columns[5])

                    # Do not add retired attributes
                    if description[0:3].upper() == "RET":
                        continue
                else:
                    description = ""

                tag = xml_element_to_tag(columns[0])
                name = xml_element_to_string(columns[1])
                keyword = xml_element_to_string(columns[2])
                vr = xml_element_to_vr(columns[3])
                vm = xml_element_to_vm(columns[4])

                attributes[tag] = Attribute(
                    tag=tag,
                    name=name,
                    keyword=keyword,
                    vr=vr,
                    vm=vm,
                    description=description
                )

                if vr != g_invalid:
                    vrs.add(vr)

                if vm != g_invalid:
                    vms.add(vm)

    return attributes, vrs, vms


def find_sibling_attribute_elements(level_string: str, root_list: list[AttributeElement]) -> list[AttributeElement]:
    """Get the nested list of data elements"""

    # Find the level
    if level_string is not None:
        level = level_string.count(">")
    else:
        level = 0

    # If we are "nested"
    if level > 0:
        # Get the right data element
        nested_attribute_element = root_list[-1]
        level -= 1

        while level > 0:
            nested_attribute_element = nested_attribute_element.attribute_elements[-1]
            level -= 1

        return nested_attribute_element.attribute_elements
    else:
        return root_list


def is_tag(element: xml.etree.ElementTree.Element) -> bool:
    """Check if an element is a tag"""
    string = xml_element_to_string(element)

    if string is None or len(string) == 0:
        return False

    return g_tag_regex.match(string) is not None


def tag_as_argument(string: str) -> Tag:
    """Return a tag as an argument"""
    match = g_tag_regex.match(string)

    if match is None:
        raise ValueError(f"Invalid tag: {string}")

    return Tag(match.group(1), match.group(2))


def sop_uid_as_argument(string: str) -> str:
    """Return a tag as an argument"""
    if string is None or len(string) == 0:
        raise ValueError("Invalid SOP UID")

    string = string.strip()

    if len(string) > 64:
        raise ValueError(f"SOP UID too long: {string}")

    match = g_uid_regex.match(string)

    if match is None:
        wildcard_match = g_uid_wildcard_regex.match(string)

        if wildcard_match is None:
            raise ValueError(f"Invalid SOP UID: {string}")

    return string


def is_attribute_type(element: xml.etree.ElementTree.Element) -> bool:
    """Check if an element is a type"""
    attribute_type = xml_element_to_string(element)

    if attribute_type is None or len(attribute_type) == 0:
        return False

    if (
            attribute_type != "1"
            and attribute_type != "1C"
            and attribute_type != "2"
            and attribute_type != "2C"
            and attribute_type != "3"
    ):
        return False

    return True


def is_module_usage(element: xml.etree.ElementTree.Element) -> bool:
    """Check if an element is an usage"""
    usage = xml_element_to_string(element)

    if usage is None or len(usage) == 0:
        return False

    if (
            usage != "M"
            and usage != "U"
            and not usage.startswith("C - ")
    ):
        return False

    return True


def fetch_attributes(table_id: str) -> list[AttributeElement]:
    """fetch all data element from link recursively"""

    if table_id in g_attribute_elements_cache:
        return g_attribute_elements_cache[table_id]

    part3 = open_document("part03")

    attribute_elements = list[AttributeElement]()

    # Find the table
    print(f"Opening {table_id}...")
    table = part3.find(f".//*[@xml:id=\"{table_id}\"]", g_namespaces)

    # For each row
    for row in table.iterfind("./docbook:tbody/docbook:tr", g_namespaces):

        # Get all columns
        columns = row.findall("./docbook:td/docbook:para", g_namespaces)

        # Get the attribute name to have the nested level and to look for special "any attribute" cases
        name = xml_element_to_string(columns[0])

        # Check for tag first because it is easier to discriminate links
        tag = None
        attribute_type = None

        # Get the tag and the type, if any.
        # Since there are "spanning", the index may vary and make the parsing more complex.
        # Thus, we simply try to guess if the current column look like a "Tag" or a type column
        for column in row.iterfind("./docbook:td/docbook:para", g_namespaces):
            if is_tag(column):
                tag = xml_element_to_tag(column)

            if is_attribute_type(column):
                attribute_type = xml_element_to_string(column)

        # Special case for "any attribute"
        if tag is None and re.match(r".*any.*attribute.*", name.lower()) is not None:
            tag = Tag("xxxx", "xxxx")

        if tag is not None:

            # Find the attribute
            attribute = g_attributes[tag]

            if len(columns) >= 4:
                description = xml_element_to_string(columns[3])
            elif len(columns) >= 3:
                description = xml_element_to_string(columns[2])
            else:
                description = ""

            if attribute_type is not None:
                # This row is most likely a data element without type
                attribute_element = AttributeElement(
                    attribute=attribute,
                    type=attribute_type,
                    description=description,
                    attribute_elements=list[AttributeElement](),
                    origin=table_id

                )
            else:
                # This row is most likely a data element without type
                attribute_element = AttributeElement(
                    attribute=attribute,
                    type=g_invalid,
                    description=description,
                    attribute_elements=list[AttributeElement](),
                    origin=table_id
                )

            # Find where we must add the new element
            sibling_attribute_elements = find_sibling_attribute_elements(name, attribute_elements)
            sibling_attribute_elements.append(attribute_element)

            # Add to the cache for later use and manage circular references
            g_attribute_elements_cache[table_id] = attribute_elements

        else:

            # Try to find a link
            xref = None
            emphasis = None

            # Do it per column because of some malformed tables where xref element is a sibling not a child of emphasis
            for column in row.iterfind("./docbook:td/docbook:para", g_namespaces):
                if xref is None:
                    xref = column.find(".//docbook:xref", g_namespaces)

                if emphasis is None:
                    emphasis = column.find(".//docbook:emphasis", g_namespaces)

            if xref is not None:
                # Get the linked table
                linked_table = xref.attrib["linkend"]

                # Retrieve the linked data elements
                linked_elements = fetch_attributes(linked_table)

                # Find where we must add the linked macro
                sibling_attribute_elements = find_sibling_attribute_elements(emphasis.text, attribute_elements)
                sibling_attribute_elements.extend(linked_elements)

                # Add to the cache for later use and manage circular references
                g_attribute_elements_cache[table_id] = attribute_elements

            # Exclusion for special cases (comments, Functional Groups, etc.)
            elif name == "BASIC CODED ENTRY ATTRIBUTES":
                continue
            elif name == "ENHANCED ENCODING MODE":
                continue
            elif len(attribute_elements) > 0 and (
                    attribute_elements[-1].attribute.tag == Tag("5200", "9229")
                    or attribute_elements[-1].attribute.tag == Tag("5200", "9230")
            ):
                # We ignore Functional Groups for now since we need to replace them by the IOD specific ones
                continue
            else:
                print(f"Unable to find xref or tag in table {table_id}")

    return attribute_elements


def fetch_module(module_section_id: str) -> Module:
    """Fetch a module"""

    # If already fetched, return it
    if module_section_id in g_modules_cache:
        return g_modules_cache[module_section_id]

    part3 = open_document("part03")

    # Find the section
    section = part3.find(f".//*[@xml:id=\"{module_section_id}\"]", g_namespaces)

    # Find the section title for the name
    name = section.find("./docbook:title", g_namespaces).text

    print(f"Fetching Module {name}...")

    attribute_elements = list[AttributeElement]()

    # Find the module attributes table
    for attributes_table in section.iterfind(".//docbook:table", g_namespaces):
        caption = attributes_table.find("./docbook:caption", g_namespaces)

        # Eliminate unwanted tables
        if caption is None:
            continue

        if caption.text is None:
            continue

        lower_caption = caption.text.lower()

        if lower_caption.find("example") != -1 or (
                lower_caption.find("module attributes") == -1
                and lower_caption.find("macro attributes") == -1
                and lower_caption.find("characteristics attributes") == -1
                and lower_caption.find("module table") == -1
        ):
            continue

        # Get the table id
        attributes_table_id = attributes_table.attrib["{http://www.w3.org/XML/1998/namespace}id"]

        # Fetch the attributes
        attribute_elements.extend(fetch_attributes(attributes_table_id))

    if len(attribute_elements) == 0:
        raise exception(f"Unable to find any attributes in {name}")

    module = Module(
        name=name,
        keyword=to_keyword(name),
        attribute_elements=attribute_elements
    )

    g_modules_cache[module_section_id] = module

    return module


def fetch_module_usages(iod_modules_table: xml.etree.ElementTree.Element) -> list[ModuleElement]:
    """Build ModuleUsage list from a modules table"""
    module_usages = list[ModuleElement]()

    # For each row get the referenced module
    for row in iod_modules_table.iterfind("./docbook:tbody/docbook:tr", g_namespaces):
        # Get the linked module
        module_link = row.find("./docbook:td/docbook:para/docbook:xref", g_namespaces)

        # Workaround for malformed XML
        if module_link is None:
            module_link = row.find("./docbook:td/docbook:xref", g_namespaces)

        module_section_id = module_link.attrib["linkend"]

        usage = "M"
        condition = ""

        # Get the usage, if any. The index may vary and make the parsing more complex
        # Thus, we simply try to guess if the current column look like an "Usage" column
        for column in row.iterfind("./docbook:td/docbook:para", g_namespaces):
            if is_module_usage(column):
                usage, condition = xml_element_to_usage(column)
                break

        module_usage = ModuleElement(
            module=fetch_module(module_section_id),
            usage=usage,
            condition=condition
        )

        module_usages.append(module_usage)

    if len(module_usages) == 0:
        raise exception(f"Unable to find any module in table.")

    return module_usages


def fetch_iod(iod_section_id: str) -> Iod:
    """Fetch an IOD from a section"""
    # Look first in the cache
    if iod_section_id in g_iods_cache:
        return g_iods_cache[iod_section_id]

    part3 = open_document("part03")

    # Find the section
    iod_section = part3.find(f".//*[@xml:id=\"{iod_section_id}\"]", g_namespaces)

    # Find the section title for the name
    name = iod_section.find("./docbook:title", g_namespaces).text

    print(f"Fetching IOD {name}...")

    modules = list[ModuleElement]()
    functional_groups = list[ModuleElement]()

    # Find the IOD modules table
    for iod_modules_table in iod_section.iterfind(".//docbook:table", g_namespaces):
        caption = iod_modules_table.find("./docbook:caption", g_namespaces)

        # Eliminate unwanted tables
        if caption is None:
            continue

        if caption.text is None:
            continue

        lower_caption = caption.text.lower()

        if lower_caption.find("example") != -1:
            continue

        if lower_caption.find("iod modules") != -1:
            # Add regular modules to the IOD
            modules.extend(fetch_module_usages(iod_modules_table))
        elif lower_caption.find("group macros") != -1:
            # The IOD section contains a table with Functional Groups, add them
            functional_groups.extend(fetch_module_usages(iod_modules_table))

    if len(modules) == 0:
        raise exception(f"Unable to find any module in {name}")

    iod = Iod(
        name=name,
        keyword=to_keyword(name),
        module_elements=modules,
        functional_group_elements=functional_groups
    )

    # Store in cache for later use
    g_iods_cache[iod_section_id] = iod

    return iod


def parse_sop_classes() -> dict[Uid, Sop]:
    """Parse SOP classes from DICOM docbook files"""
    sop_classes = dict[Uid, Sop]()

    part4 = open_document("part04")

    # For each tables
    for table_id in {
        "table_B.5-1",
        "table_I.4-1",
        "table_GG.3-1",
        "table_KK.1-2"
    }:
        # Find the table
        table = part4.find(f".//*[@xml:id=\"{table_id}\"]", g_namespaces)

        # For each rows
        for row in table.iterfind("./docbook:tbody/docbook:tr", g_namespaces):

            # Read all columns
            columns = row.findall("./docbook:td/docbook:para", g_namespaces)

            if len(columns) >= 3:

                uid = g_uids[xml_element_to_string(columns[1])]

                # Here we have to look if we need to filter this uid
                if g_sop_list is not None:

                    keep = False

                    for sop_pattern in g_sop_list:
                        if fnmatch.fnmatch(uid.value, sop_pattern):
                            keep = True

                    if not keep:
                        continue

                # Column 3 is a link to IOD in part 3, get it
                section_link = columns[2].find("./docbook:olink", g_namespaces)

                # Workaround for malformed XML
                if section_link is None:
                    section_link = row.find("./docbook:td/docbook:olink", g_namespaces)

                # Follow the link...
                section_id = section_link.attrib["targetptr"]
                iod = fetch_iod(section_id)

                sop = Sop(
                    uid=uid,
                    iod=iod
                )

                # Check if all tags is present
                if check_mandatory_tags(sop):
                    print(f"Adding SOP {sop.uid.name}...")
                    sop_classes[sop.uid] = sop

    return sop_classes


def parse_uids() -> tuple[dict[str, Uid], set[str]]:
    """Parse UIDs from DICOM docbook files"""
    uids = dict[str, Uid]()
    uid_types = set[str]()

    part6 = open_document("part06")

    # Find the section
    uids_table = part6.find(f".//*[@xml:id=\"table_A-1\"]", g_namespaces)
    for row in uids_table.iterfind("./docbook:tbody/docbook:tr", g_namespaces):
        columns = row.findall("./docbook:td/docbook:para", g_namespaces)

        name = xml_element_to_string(columns[1])

        # Do not add retired UIDs
        if name.lower().find("retired") != -1:
            continue

        uid_type = xml_element_to_string(columns[3])
        uid_types.add(uid_type)

        uid = Uid(
            value=xml_element_to_uid(columns[0]),
            name=name,
            keyword=xml_element_to_string(columns[2]),
            type=uid_type
        )

        uids[uid.value] = uid

    return uids, uid_types


def print_sop(sop: Sop):
    """Print a SOP class"""
    print(f"{sop.uid.name}")
    print(f"├─UID: {sop.uid.value}")
    print(f"├─Keyword: {sop.uid.keyword}")
    print(f"└─{sop.iod.name}")
    print(f"  ├─IOD keyword: {sop.iod.keyword}")
    print(f"  └─IOD modules ({len(sop.iod.module_elements)}):")

    for module_usage_index, module_usage in enumerate(sop.iod.module_elements):

        line = f"{module_usage.module.name} ({module_usage.usage}):"

        is_last_module = module_usage_index >= len(sop.iod.module_elements) - 1
        if not is_last_module:
            print(f"    ├─{line}")
        else:
            print(f"    └─{line}")

        def print_attribute_elements(tree: str, attribute_elements: list[AttributeElement], path: list[str]):

            for attribute_element_index, attribute_element in enumerate(attribute_elements):

                is_functional_group = attribute_element.attribute.tag == Tag(
                    group="5200",
                    element="9229"
                ) or attribute_element.attribute.tag == Tag(
                    group="5200",
                    element="9230"
                )

                attribute_element_line = f"({attribute_element.attribute.tag.group}, {attribute_element.attribute.tag.element}) ({attribute_element.type}) {attribute_element.attribute.name}"

                if attribute_element.attribute_elements is not None and len(attribute_element.attribute_elements) > 0:
                    attribute_element_line += f" ({len(attribute_element.attribute_elements)}):"

                is_last_element = attribute_element_index >= len(attribute_elements) - 1
                if not is_last_element:
                    print(tree + f"├─{attribute_element_line}")
                else:
                    print(tree + f"└─{attribute_element_line}")

                if (
                        attribute_element.attribute_elements is not None
                        and len(attribute_element.attribute_elements) > 0
                        and attribute_element.origin not in path
                ):
                    # This is for breaking circular references
                    new_path = path + [attribute_element.origin]

                    if is_last_element and not is_functional_group:
                        print_attribute_elements(tree + "   ", attribute_element.attribute_elements, new_path)
                    else:
                        print_attribute_elements(tree + "│  ", attribute_element.attribute_elements, new_path)

                if is_functional_group:
                    if is_last_element:
                        tree += "   "
                    else:
                        tree += "│  "

                    for functional_groups_index, functional_groups_usage in enumerate(
                            sop.iod.functional_group_elements):
                        functional_line = f"{functional_groups_usage.module.name} ({functional_groups_usage.usage}):"
                        if functional_groups_index < len(sop.iod.functional_group_elements) - 1:
                            print(tree + f"├─{functional_line}")
                        else:
                            print(tree + f"└─{functional_line}")

                        if (
                                functional_groups_usage.module.attribute_elements is not None
                                and len(functional_groups_usage.module.attribute_elements) > 0
                        ):

                            # This is for breaking circular references
                            new_path = path + [attribute_element.origin]

                            if is_last_element:
                                print_attribute_elements(
                                    tree + "   ", functional_groups_usage.module.attribute_elements, new_path
                                )
                            else:
                                print_attribute_elements(
                                    tree + "│  ", functional_groups_usage.module.attribute_elements, new_path
                                )

        if is_last_module:
            print_attribute_elements("       ", module_usage.module.attribute_elements, [])
        else:
            print_attribute_elements("    │  ", module_usage.module.attribute_elements, [])

    print()


def main():
    """Main function"""

    # Parse arguments
    global g_parser
    g_parser = argparse.ArgumentParser("Generate DICOM dictionary files")

    g_parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print verbose output"
    )

    g_parser.add_argument(
        "source_url",
        nargs="?",
        default="http://dicom.nema.org/medical/dicom/current/source/docbook",
        help="Source URL of the DICOM docbook files"
    )

    g_parser.add_argument(
        "--sop-list",
        nargs="+",
        help="Only search for given SOP classes",
        dest="sop_list",
        type=sop_uid_as_argument
    )

    g_parser.add_argument(
        "--mandatory-tags",
        nargs="+",
        help="Only add SOP classes which contain the given tags",
        dest="mandatory_tags",
        type=tag_as_argument
    )

    global g_arguments
    g_arguments = g_parser.parse_args()

    global g_sop_list
    g_sop_list = g_arguments.sop_list

    global g_mandatory_tags
    g_mandatory_tags = g_arguments.mandatory_tags

    global g_source_url
    g_source_url = g_arguments.source_url

    # Parse Attributes
    global g_attributes, g_vrs, g_vms
    g_attributes, g_vrs, g_vms = parse_attributes()

    # Add Special case "any" attribute
    g_attributes[Tag("xxxx", "xxxx")] = Attribute(
        tag=Tag("xxxx", "xxxx"),
        name="Any",
        keyword="Any",
        vr="UNDEFINED",
        vm="UNDEFINED",
        description=""
    )

    # Parse UIDs
    global g_uids, g_uid_types
    g_uids, g_uid_types = parse_uids()

    # Parse SOPs
    global g_filtered_sops
    g_filtered_sops = parse_sop_classes()

    # filter results
    global g_filtered_iods
    global g_filtered_modules
    global g_filtered_attributes
    global g_filtered_attribute_types

    for sop in g_filtered_sops.values():

        # Add used IODs
        g_filtered_iods[sop.iod.keyword] = sop.iod

        # Recursively add used attributes
        def walk_for_attributes(attribute_elements: list[AttributeElement], path: list[str]):
            for attribute_element in attribute_elements:

                # Add used attribute
                g_filtered_attributes[attribute_element.attribute.tag] = attribute_element.attribute

                # Add used attribute type
                g_filtered_attribute_types.add(attribute_element.type)

                # Recurse through sub-elements
                if (
                        attribute_element.attribute_elements is not None
                        and len(attribute_element.attribute_elements) > 0
                        and attribute_element.origin not in path
                ):
                    # The "path" allows breaking circular references
                    new_path = path + [attribute_element.origin]
                    walk_for_attributes(attribute_element.attribute_elements, new_path)

        for module_element in sop.iod.module_elements:
            # Add used module
            g_filtered_modules[module_element.module.keyword] = module_element.module

            # Walk recursively through data elements
            walk_for_attributes(module_element.module.attribute_elements, [])

        if sop.iod.functional_group_elements is not None:
            for module_element in sop.iod.functional_group_elements:
                # Add used module
                g_filtered_modules[module_element.module.keyword] = module_element.module

                # Walk recursively through data elements
                walk_for_attributes(module_element.module.attribute_elements, [])

    print("\n\nParsing results:")
    print("_" * 120)

    # Display results:
    if g_sop_list is not None:
        print(f"Filtering by SOP Class UID: {str(g_sop_list)}.")

    if g_mandatory_tags is not None:
        print(f"Filtering by mandatory tags: {str(g_mandatory_tags)}.")

    print("_" * 120)

    print(f"\nFound {len(g_filtered_sops)} SOP classes:")
    for sop in g_filtered_sops.values():
        print(f"\t{sop.uid.value}: {sop.uid.name}")

    print("_" * 120)

    print(f"\nFound {len(g_filtered_iods)} IODs:")
    for iod in g_filtered_iods.values():
        print(f"\t{iod.name}")

    print("_" * 120)

    print(f"\nFound {len(g_filtered_modules)} Modules:")
    for module in g_filtered_modules.values():
        print(f"\t{module.name}")

    print("_" * 120)

    print(f"\nFound {len(g_filtered_attributes)} Attributes:")
    for attribute in g_filtered_attributes.values():
        print(f"\t({attribute.tag.group}, {attribute.tag.element}): {attribute.name}")

    print("_" * 120)

    if g_arguments.verbose:
        print(f"\nSOP classes details:\n")

        # Print SOP classes details
        for sop in g_filtered_sops.values():
            print_sop(sop)
            print("_" * 120)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#! /usr/bin/env python

from jinja2 import Environment, PackageLoader, Template
from pprint import pprint
import copy

def appendDict(dict1, dict2):
    res = { k: v[:] for k,v in dict1.items() }
    for k in dict2.keys():
        if k in dict1:
            res[k] += dict2[k][:]
        else:
            res[k] = dict2[k][:]
    return res

## Function that generate permutations of the configurations in parameters/
## It takes a base config A and combine it with all the subsequent configurations 1, 2, 3,...
## It will thus generate for instance [A,1] [A,2], [A,1,2], [A,3], [A,1,3], [A,2,3], [A,1,2,3]
def generatePermutations(baseConfig, *configs):
    permutations = [baseConfig]

    # For each configuration
    for currentConfig in configs:

        # Iterate over all permutations generated before the currentConfig
        # At the beginning, only baseConfig, then growing with the permutations we create
        newPermutations  = []

        for config in permutations:

            # Concatenate name
            currentPermutation  = []
            currentPermutation += [ config[0] + '+' + currentConfig[0] ]

            # Concatenate defines
            if config[1]:
                if currentConfig[1]:
                    currentPermutation += [config[1] + ',' + currentConfig[1]]
                else:
                    currentPermutation += [config[1]]
            else:
                if currentConfig[1]:
                    currentPermutation += [currentConfig[1]]

            # Concatenate params
            paramsDict = appendDict(config[2], currentConfig[2])

            currentPermutation.append(paramsDict)

            newPermutations += [currentPermutation]

        permutations += newPermutations

    return permutations


###############################################################################
## Shader Parameters
###############################################################################

## Common parameters used in Lighting.glsl, used both at Vertex and Fragment stages
lightingParams = [ '// Lighting',
                   'param_named_auto u_f3CameraPos camera_position',
                   'param_named u_fNumLights float 1.0',
                   'param_named_auto u_f4LightAmbientCol ambient_light_colour',
                   'param_named_auto u_f3LightDiffuseCol light_diffuse_colour_array 8',
                   'param_named_auto u_f3LightSpecularCol light_specular_colour_array 8',
                   'param_named_auto u_f4AmbientCol surface_ambient_colour',
                   'param_named_auto u_f4DiffuseCol surface_diffuse_colour',
                   'param_named_auto u_f4SpecularCol surface_specular_colour',
                   'param_named_auto u_fShininess surface_shininess']

lightingParamsFlat = lightingParams + [ 'param_named_auto u_f4LightPosVs light_position_view_space_array 8' ]

lightingParams += [ 'param_named_auto u_f4LightPos light_position_array 10' ]

## Common parameters used for ambient lighting
ambientParams = [ '// Ambient',
                  'param_named_auto u_ambient surface_ambient_colour',
                  'param_named_auto u_diffuse surface_diffuse_colour',
                  'param_named_auto u_f4LightAmbientCol ambient_light_colour']

## Diffuse or negato textures parameters, they should be bound at different unit depending on OIT technique
texParams = ['// Diffuse texture', 'param_named u_texture int 0']

texAlphaParams = ['// Use alpha channel from the texture', 'param_named u_useTextureAlpha int 0']

## Parameters used for Negato.glsl
negatoParams = [ '// Negato',
                 'param_named u_minValue float 0.0',
                 'param_named u_maxValue float 1.0',
                 'param_named u_slice float 0',
                 'param_named u_orientation int 2']

## Diffuse needed alone in some cases
diffuseColorParams = ['// Diffuse color',
                      'param_named_auto u_diffuse surface_diffuse_colour']

## Per primitive color
ppColorParams = ['param_named u_colorPrimitiveTexture int 0']

vpFlatParams = ['param_named_auto u_worldView worldview_matrix']
vpPixelLitParams = ['param_named_auto u_world world_matrix',
                    'param_named_auto u_normalMatrix inverse_transpose_world_matrix']

## 'Name', '#Define', 'Attached vp', {parameters dict}
cfgAmbient = ['Ambient', '', { 'renderSceneVP' : ambientParams } ]

cfgFlat = ['Flat', 'FLAT=1', { 'renderSceneVP' : vpFlatParams, 'defaultFP' : lightingParamsFlat } ]

cfgPixelLit = ['PixelLit', 'PHONG=1', { 'renderSceneVP' : vpPixelLitParams,
                                        'defaultFP' : lightingParams,
                                        'depthPeelingFP' : lightingParams,
                                        'dualDepthPeelingFP' : lightingParams,
                                        'HT_weight_blendFP' : lightingParams,
                                        'weighted_blendFP' : lightingParams } ]

cfgEdgeNormal = ['Edge_Normal', 'EDGE_NORMAL=1', { } ]

cfgVertexColor = ['VT', 'VERTEX_COLOR=1', { } ]

cfgDiffuseTex = ['DfsTex', 'DIFFUSE_TEX=1', { 'defaultFP' : texParams + texAlphaParams,
                                                          'depthPeelingFP' : texParams + texAlphaParams,
                                                          'dualDepthPeelingFP' : texParams + texAlphaParams,
                                                          'HT_weight_blendFP' : texParams + texAlphaParams,
                                                          'weighted_blendFP' : texParams + texAlphaParams } ]

cfgTriangles = ['Triangles', 'TRIANGLES=1', { 'renderSceneGP' : [], } ]
cfgQuad = ['Quad', 'QUAD=1', { } ]
cfgTetra = ['Tetra', 'TETRA=1', { } ]

cfgPerPrimitiveColor = ['PPColor', 'PER_PRIMITIVE_COLOR=1', { 'renderSceneGP' : ppColorParams} ]

## Configurations for vertex programs
## Basis are the different lighting techniques, and optional are vertex color and diffuse texture
configsListVP = []

configsListVP += generatePermutations(cfgAmbient, cfgVertexColor, cfgDiffuseTex)
configsListVP += generatePermutations(cfgFlat, cfgVertexColor, cfgDiffuseTex)
configsListVP += generatePermutations(cfgPixelLit, cfgVertexColor, cfgDiffuseTex)

## Configurations for fragment programs
## Base are the different lighting techniques, and optional are vertex color and diffuse texture
configsListFP = []

configsListFP += generatePermutations(cfgAmbient, cfgVertexColor, cfgDiffuseTex)
configsListFP += generatePermutations(cfgFlat, cfgVertexColor, cfgDiffuseTex)
configsListFP += generatePermutations(cfgPixelLit, cfgVertexColor, cfgDiffuseTex)

configsListFP += [cfgEdgeNormal]

## Configurations for geometry programs
## Base are the different primitives types to generate, and optional are vertex color, diffuse texture and per-primitive
## color
configsListGP = []

configsListGP += generatePermutations(cfgTriangles, cfgVertexColor, cfgDiffuseTex, cfgPerPrimitiveColor)
configsListGP += generatePermutations(cfgQuad, cfgVertexColor, cfgDiffuseTex, cfgPerPrimitiveColor)
configsListGP += generatePermutations(cfgTetra, cfgVertexColor, cfgDiffuseTex, cfgPerPrimitiveColor)

#pprint(configsListVP)

env = Environment(loader=PackageLoader('genDefaultMaterial', 'genTemplates'),trim_blocks=True)
template = env.get_template('Default.program.tpl')
template.stream(configsVP=configsListVP, configsFP=configsListFP).dump('scene/Default.program')

template = env.get_template('R2VB.program.tpl')
template.stream(configsVP=configsListVP, configsGP=configsListGP).dump('core/R2VB.program')
